# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Explorer Plugin API.
"""

# Local imports
from spyder.plugins.explorer.widgets.explorer import (
    DirViewActions, DirViewContextMenuSections, DirViewMenus,
    DirViewNewSubMenuSections, ExplorerTreeWidgetActions)

from spyder.plugins.explorer.widgets.main_widget import (
    ExplorerWidgetMainToolbarSections, ExplorerWidgetOptionsMenuSections)
